#version 330

struct Light {
	vec3 lightColor;
	float ambientIntensity;
	float diffuseIntensity;
};

struct DirectionalLight {
	Light light;
	vec3 lightDirection;
};

struct PointLight {
	Light light;
	vec3 lightPosition;
	float attConst;
	float attLin;
	float attExp;
};

struct SpotLight {
	PointLight light;
	vec3 lightDirection;
	float cut;
};


in vec3 worldNormal;
in vec3 worldPos;
in vec2 fragUV;
in vec3 worldTangent;
in vec4 lightPos;


uniform DirectionalLight dl;

uniform int pointLightCount;
uniform PointLight pointLights[2];

uniform int spotLightCount;
uniform SpotLight spotLights[1];

uniform vec3 viewWorldPos;
uniform float specularIntensity;
uniform float specularPower;


uniform sampler2D colorTexture;
uniform sampler2D normalMap;
uniform sampler2D shadowMap;
//uniform sampler2DShadow shadowMap;


out vec4 fragColor;


float createShadows(vec4 lightPos){
	//vec3 tmp = vec3((0.5f * lightPos.x + 0.5f), (0.5f * lightPos.y + 0.5f), (0.5f * lightPos.z + 0.5f));
	//vec4 texCoords = vec4(tmp, lightPos.w);
	//float depth = textureProj(shadowMap, texCoords, 0.00001f);
	//return depth;

    vec3 texCoords = lightPos.xyz / lightPos.w;
    vec2 uvCoords;
    uvCoords.x = 0.5f * texCoords.x + 0.5f;
    uvCoords.y = 0.5f * texCoords.y + 0.5f;
    float z = 0.5f * texCoords.z + 0.5f;
    float depth = texture(shadowMap, uvCoords).x;
    if(depth < z + 0.00001f){
        return 0.01f;
    } else {
        return 1.0f;
	}
}

vec4 createLight(Light l, vec3 lightDirection, vec3 normal, float shadowMultiplier){
	vec4 ambientColor = vec4(l.lightColor, 1.0f) * l.ambientIntensity;
	float diffuseFactor = dot(normal, -lightDirection);

	vec4 diffuseColor = vec4(0.0f, 0.0f, 0.0f, 0.0f);
	vec4 specularColor = vec4(0.0f, 0.0f, 0.0f, 0.0f);

	if(diffuseFactor > 0.0f){
		diffuseColor = vec4(l.lightColor, 1.0f) * l.diffuseIntensity * diffuseFactor;

		vec3 toViewer = normalize(viewWorldPos - worldPos);
		vec3 reflected = normalize(reflect(lightDirection, normal));
		float specularFactor = pow(dot(toViewer, reflected), specularPower);
		if(specularFactor > 0){
			specularColor = vec4(l.lightColor, 1.0f) * specularIntensity * specularFactor;
		}

	}

	return (ambientColor + (diffuseColor + specularColor) * shadowMultiplier);
}

vec4 createDirectionalLight(vec3 normal){
	return createLight(dl.light, dl.lightDirection, normal, 1.0f);
}

vec4 createPointLight(PointLight l, vec3 normal, vec4 lightPos){
	vec3 lightDirection = worldPos - l.lightPosition;
    float dist = length(lightDirection);
    lightDirection = normalize(lightDirection);
	float shadowMultiplier = createShadows(lightPos);
    vec4 color = createLight(l.light, lightDirection, normal, shadowMultiplier);
    float att =  l.attConst + l.attLin * dist + l.attExp * dist * dist;
	
    return (color / att); 
}

vec4 createSpotLight(SpotLight l, vec3 normal, vec4 lightPos){
	vec3 lightToPixel = normalize(worldPos - l.light.lightPosition);
	float spotFactor = dot(lightToPixel, l.lightDirection);

	if(spotFactor > l.cut){
		vec4 color = createPointLight(l.light, normal, lightPos);
		return (color * (1.0f - (1.0f - spotFactor) * 1.0f / (1.0f - l.cut)));
	} else {
		return vec4(0.0f, 0.0f, 0.0f, 0.0f);
	}
}


vec3 createNormalMap(){
	vec3 n = normalize(worldNormal);
	vec3 t = normalize(worldTangent);
	t = normalize(t - dot(t, n) * n);
	vec3 b = cross(t, n);

	mat3 tbn = mat3(t, b, n);

	vec3 mapNormal = ((2.0f * (texture(normalMap, fragUV).xyz)) - vec3(1.0f, 1.0f, 1.0f));


	return normalize(tbn * mapNormal);
}


void main(){
	vec3 nNormal = createNormalMap();

	vec4 lightColor = createDirectionalLight(nNormal);
	for(int i = 0; i < pointLightCount; i++){
		lightColor += createPointLight(pointLights[i], nNormal, lightPos);
	}
	for(int i = 0; i < spotLightCount; i++){
		lightColor += createSpotLight(spotLights[i], nNormal, lightPos);
	}

	vec3 textureColor = texture(colorTexture, fragUV).rgb;

	fragColor = vec4(textureColor, 1.0f) * lightColor; 
}